@@ -262,9 +262,9 @@ SEXP dev_string_widths(SEXP strings, SEXP family, SEXP face, SEXP size, SEXP cex
   pGEDevDesc dev = GEcurrentDevice();
   R_GE_gcontext gc = {};
   double width = 0;
-  int n_total = LENGTH(strings);
-  int scalar_family = LENGTH(family) == 1;
-  int scalar_rest = LENGTH(face) == 1;
+  int n_total = Rf_length(strings);
+  int scalar_family = Rf_length(family) == 1;
+  int scalar_rest = Rf_length(face) == 1;
   strcpy(gc.fontfamily, Rf_translateCharUTF8(STRING_ELT(family, 0)));
   gc.fontface = INTEGER(face)[0];
   gc.ps = REAL(size)[0];
@@ -312,9 +312,9 @@ SEXP dev_string_metrics(SEXP strings, SEXP family, SEXP face, SEXP size, SEXP ce
   pGEDevDesc dev = GEcurrentDevice();
   R_GE_gcontext gc = {};
   double width = 0, ascent = 0, descent = 0;
-  int n_total = LENGTH(strings);
-  int scalar_family = LENGTH(family) == 1;
-  int scalar_rest = LENGTH(face) == 1;
+  int n_total = Rf_length(strings);
+  int scalar_family = Rf_length(family) == 1;
+  int scalar_rest = Rf_length(face) == 1;
   strcpy(gc.fontfamily, Rf_translateCharUTF8(STRING_ELT(family, 0)));
   gc.fontface = INTEGER(face)[0];
   gc.ps = REAL(size)[0];
@@ -376,7 +376,7 @@ SEXP register_font(SEXP family, SEXP paths, SEXP indices) {
   FontReg& registry = get_font_registry();
   std::string name = Rf_translateCharUTF8(STRING_ELT(family, 0));
   FontCollection col = {};
-  for (int i = 0; i < LENGTH(paths); ++i) {
+  for (int i = 0; i < Rf_length(paths); ++i) {
     std::string font_path = Rf_translateCharUTF8(STRING_ELT(paths, i));
     FontLoc font(font_path, INTEGER(indices)[i]);
     col.push_back(font);
